from datetime import (
    date,
    datetime,
    time,
    timedelta,
)
from functools import (
    partial,
)
from typing import (
    Optional,
)

from django.db.models import (
    Q,
)
from m3_ext.ui.icons import (
    Icons,
)
from objectpack.actions import (
    ObjectPack,
)

from educommon.objectpack.actions import (
    ExtObjectRowsAction,
)
from educommon.objectpack.filters import (
    BoolChoicesFilter,
    ColumnFilterEngine,
    DateFilterByAnnotatedField,
)
from educommon.utils.object_grid import (
    add_one_row_button,
    boolean_column_renderer,
)
from educommon.utils.ui import (
    append_template_globals,
)

from edu_rdm_integration.models import (
    UploaderClientLog,
)
from edu_rdm_integration.uploader_log.ui import (
    UploaderLogInfoWindow,
)


class UploaderLogPack(ObjectPack):
    """
    Пак журнала Загрузчика данных в витрину.
    """

    title = 'Журнал логов РВД'
    model = UploaderClientLog

    edit_window = UploaderLogInfoWindow

    _is_primary_for_model = False
    can_delete = False
    list_sort_order = ['-request_datetime']
    _DEFAULT_PAGING_LIMIT = 35

    filter_engine_clz = ColumnFilterEngine
    _fd = partial(DateFilterByAnnotatedField, model)

    need_check_permission = True

    columns = [
        {
            'header': '№ п/п',
            'data_index': 'row_number',
            'sortable': False,
            'width': 3,
        },
        {
            'header': 'Метод',
            'data_index': 'http_method',
            'sortable': False,
            'width': 3,
        },
        {
            'header': 'URL',
            'data_index': 'request_url',
            'searchable': True,
            'sortable': False,
            'width': 21,
        },
        {
            'header': 'Дата и время запроса',
            'data_index': 'request_datetime',
            'sortable': True,
            'filter': _fd(
                'request_datetime',
                tooltip='C',
                default_value=date.today() - timedelta(days=1),
                editable=False,
                lookup=lambda v: Q(request_datetime__gte=datetime.combine(v, time.min))
            ) & _fd(
                'request_datetime',
                tooltip='По',
                default_value=date.today,
                editable=False,
                lookup=lambda v: Q(request_datetime__lte=datetime.combine(v, time.max))
            ),
            'width': 11,
        },
        {
            'header': 'Результат',
            'data_index': 'http_response_status',
            'sortable': False,
            'width': 5,
        },
        {
            'header': 'Путь до файлов',
            'data_index': 'attachment_file',
            'sortable': False,
            'searchable': True,
            'width': 37,
        },
        {
            'header': 'ID запроса',
            'data_index': 'request_id',
            'sortable': False,
            'searchable': True,
            'width': 10,
        },
        {
            'header': 'Код статуса загрузки',
            'data_index': 'status_code',
            'sortable': False,
            'searchable': True,
            'width': 10,
        },
        {
            'header': 'Описание статуса загрузки',
            'data_index': 'status_description',
            'sortable': False,
            'searchable': True,
            'width': 10,
        },
        {
            'header': 'Режим эмуляции',
            'data_index': 'is_emulation',
            'sortable': False,
            'filter': BoolChoicesFilter('is_emulation'),
            'column_renderer': boolean_column_renderer(),
            'width': 5,
        }
    ]

    def __init__(self):
        """Инициализация."""
        super(UploaderLogPack, self).__init__()

        self.edit_window_action.perm_code = 'view'
        self.replace_action('rows_action', ExtObjectRowsAction())

    def get_rows_query(self, request, context):
        """Возвращает кварисет для получения списка данных."""
        context._row_number = int(getattr(context, 'start', 0))

        return super().get_rows_query(request, context)

    def get_row(self, row_id: Optional[int]) -> 'UploaderClientLog':
        """Возвращает объект по идентификатору."""
        if row_id:
            record = self.model.base_objects.get(id=row_id)
        else:
            record = super().get_row(row_id)

        return record

    def prepare_row(self, obj, request, context):
        """Установка дополнительных атрибутов объекта."""
        context._row_number += 1
        obj.row_number = context._row_number

        return obj

    def create_list_window(self, is_select_mode, request, context):
        """Создание окна списка."""
        win = super().create_list_window(is_select_mode, request, context)

        append_template_globals(win, 'ui-js/object-grid-buttons.js')

        return win

    def get_list_window_params(self, params, request, context):
        """Установка параметров окна списка."""
        params = super().get_list_window_params(params, request, context)

        params.update(
            maximized=True,
        )

        return params

    def configure_grid(self, grid, *args, **kwargs):
        """Конфигурирование грида окна списка."""
        super().configure_grid(grid, *args, **kwargs)

        grid.url_edit = None
        add_one_row_button(
            'Просмотр',
            grid,
            self.edit_window_action,
            Icons.APPLICATION_VIEW_DETAIL,
            dbl_clicked=True,
            index=0,
        )

    def get_edit_window_params(self, params, request, context):
        """Параметры окна редактирования."""
        params = super().get_edit_window_params(params, request, context)

        params.update(
            title=f'{self.title}: Просмотр',
        )

        return params

    def extend_menu(self, menu):  # noqa D102
        return menu.SubMenu(
            'Администрирование', menu.SubMenu(
                'Региональная витрина данных', menu.Item(
                    self.title, self.list_window_action
                )
            )
        )



from django.conf import (
    settings,
)

from educommon import (
    logger,
)
from educommon.utils.seqtools import (
    make_chunks,
)

from edu_rdm_integration.adapters.runners import (
    WebEduRunner,
)
from edu_rdm_integration.consts import (
    LOGS_DELIMITER,
)


class BaseCollectingDataRunner(WebEduRunner):
    """
    Базовый класс ранеров функций сбора данных для интеграции с "Региональная витрина данных".

    Поддерживается режим принудительного запуска функций без постановки в очередь на исполнение.
    """

    def __init__(self, *args, is_force_run: bool = False, **kwargs):
        self._is_force_run = is_force_run

        super().__init__(*args, **kwargs)

    def _populate_queue_by_runnable_classes(self, logs, *args, **kwargs):
        """
        Заполнение очереди запускаемыми объектами.
        """
        raw_logs_chunks = make_chunks(
            iterable=logs,
            size=settings.RDM_COLLECT_CHUNK_SIZE,
        )

        for chunk_index, raw_logs in enumerate(raw_logs_chunks, start=1):
            runnable_classes = self._prepare_runnable_classes()

            raw_logs = list(raw_logs)

            for runnable_class in runnable_classes:
                runnable = runnable_class(raw_logs=raw_logs, *args, **kwargs)

                # TODO EDUSCHL-20274 Реализация форсированного выполнения функций
                if self._is_force_run:
                    logger.info(
                        f'{LOGS_DELIMITER * 2}force run {runnable_class.__name__} with logs chunk {chunk_index}..'
                    )
                    self.force_run(runnable=runnable, *args, **kwargs)
                else:
                    logger.info(
                        f'{LOGS_DELIMITER * 2}enqueue {runnable_class.__name__} with logs chunk {chunk_index}..'
                    )

                    self.enqueue(runnable=runnable, *args, **kwargs)

    # TODO EDUSCHL-20274 Реализация форсированного выполнения функций
    def force_run(self, runnable, *args, **kwargs):
        """Принудительное выполнение запускаемого объекта."""
        self.before_validate()
        self.validate()
        self.after_validate()

        if self.result.has_not_errors:
            runnable.before_validate()
            runnable.validate()
            runnable.after_validate()

            runnable.before_run(*args, **kwargs)
            runnable.run(*args, **kwargs)
            runnable.after_run(*args, **kwargs)

            self.result.append_entity(runnable.result)



import uuid
from datetime import (
    datetime,
)
from typing import (
    List,
)

import django.utils.timezone
from django.contrib.postgres.fields import (
    JSONField,
)
from django.db.models import (
    CASCADE,
    PROTECT,
    SET_NULL,
    BooleanField,
    CharField,
    DateTimeField,
    FileField,
    ForeignKey,
    SmallIntegerField,
    UUIDField,
)
from m3.db import (
    BaseObjectModel,
)
from uploader_client.models import (
    Entry,
)

from educommon.django.db.mixins import (
    ReprStrPreModelMixin,
)
from educommon.integration_entities.enums import (
    EntityLogOperation,
)
from educommon.utils.caching import (
    cached_function,
)
from educommon.utils.date import (
    get_today_max_datetime,
    get_today_min_datetime,
)
from function_tools.models import (
    Entity,
)
from m3_db_utils.models import (
    ModelEnumValue,
    TitledIntegerModelEnum,
    TitledModelEnum,
)

from edu_rdm_integration.enums import (
    FileUploadStatusEnum,
)
from edu_rdm_integration.utils import (
    get_exporting_data_stage_attachment_path,
)
from io import (
    StringIO,
)
from typing import (
    Optional,
    Tuple,
)

from django.db.models import (
    Manager,
)
from django.utils.functional import (
    cached_property,
)

from uploader_client.models import (
    Entry,
)

from edu_rdm_integration.uploader_log.managers import (
    UploaderClientLogManager,
)


class CollectingDataStageStatus(TitledModelEnum):
    """Статус этапа сбора данных."""

    CREATED = ModelEnumValue(
        title='Создан',
    )

    IN_PROGRESS = ModelEnumValue(
        title='В процессе сбора',
    )

    FAILED = ModelEnumValue(
        title='Завершено с ошибками',
    )

    FINISHED = ModelEnumValue(
        title='Завершено',
    )

    class Meta:
        db_table = 'rdm_collecting_data_stage_status'
        verbose_name = 'Модель-перечисление статусов этапа сбора данных'
        verbose_name_plural = 'Модели-перечисления статусов этапов сбора данных'


class CollectingExportedDataStage(ReprStrPreModelMixin, BaseObjectModel):
    """Этап подготовки данных в рамках Функций. За работу Функции отвечает ранер менеджер."""

    manager = ForeignKey(
        to=Entity,
        verbose_name='Менеджер ранера Функции',
        on_delete=PROTECT,
        null=True,
        blank=True,
    )

    logs_period_started_at = DateTimeField(
        'Левая граница периода обрабатываемых логов',
        db_index=True,
        default=get_today_min_datetime,
    )

    logs_period_ended_at = DateTimeField(
        'Правая граница периода обрабатываемых логов',
        db_index=True,
        default=get_today_max_datetime,
    )

    started_at = DateTimeField(
        'Время начала сбора данных',
        auto_now_add=True,
        db_index=True,
    )

    ended_at = DateTimeField(
        'Время завершения сбора данных',
        null=True,
        blank=True,
        db_index=True,
    )

    status = ForeignKey(
        to=CollectingDataStageStatus,
        verbose_name='Статус',
        on_delete=PROTECT,
        default=CollectingDataStageStatus.CREATED.key,
    )

    class Meta:
        db_table = 'rdm_collecting_exported_data_stage'
        verbose_name = 'Этап формирования данных для выгрузки'
        verbose_name_plural = 'Этапы формирования данных для выгрузки'

    @property
    def attrs_for_repr_str(self):
        return ['manager_id', 'logs_period_started_at', 'logs_period_ended_at', 'started_at', 'ended_at', 'status_id']

    def save(self, *args, **kwargs):
        if (
            self.status_id in (CollectingDataStageStatus.FAILED.key, CollectingDataStageStatus.FINISHED.key)
            and not self.ended_at
        ):
            self.ended_at = datetime.now()

        super().save(*args, **kwargs)


class CollectingDataSubStageStatus(TitledModelEnum):
    """Статус этапа сбора данных."""

    CREATED = ModelEnumValue(
        title='Создан',
    )

    IN_PROGRESS = ModelEnumValue(
        title='В процессе сбора',
    )

    READY_TO_EXPORT = ModelEnumValue(
        title='Готово к выгрузке',
    )

    FAILED = ModelEnumValue(
        title='Завершено с ошибками',
    )

    EXPORTED = ModelEnumValue(
        title='Выгружено',
    )

    NOT_EXPORTED = ModelEnumValue(
        title='Не выгружено',
    )

    class Meta:
        db_table = 'rdm_collecting_data_sub_stage_status'
        verbose_name = 'Модель-перечисление статусов подэтапа сбора данных'
        verbose_name_plural = 'Модели-перечисления статусов подэтапов сбора данных'


class CollectingExportedDataSubStage(ReprStrPreModelMixin, BaseObjectModel):
    """Подэтап сбора данных для сущностей в рамках функции."""

    stage = ForeignKey(
        to=CollectingExportedDataStage,
        verbose_name='Этап подготовки данных для экспорта',
        on_delete=PROTECT,
    )

    function = ForeignKey(
        to=Entity,
        verbose_name='Функция',
        on_delete=PROTECT,
    )

    started_at = DateTimeField(
        'Время начала сбора данных',
        auto_now_add=True,
        db_index=True,
    )

    ended_at = DateTimeField(
        'Время завершения сбора данных',
        null=True,
        blank=True,
        db_index=True,
    )

    previous = ForeignKey(
        'self',
        null=True,
        blank=True,
        verbose_name='Предыдущий сбор данных',
        on_delete=CASCADE,
    )

    status = ForeignKey(
        to=CollectingDataSubStageStatus,
        verbose_name='Статус',
        on_delete=PROTECT,
        default=CollectingDataSubStageStatus.CREATED.key,
    )

    class Meta:
        db_table = 'rdm_collecting_exported_data_sub_stage'
        verbose_name = 'Подэтап формирования данных для выгрузки'
        verbose_name_plural = 'Подэтапы формирования данных для выгрузки'

    @property
    def attrs_for_repr_str(self):
        return ['stage_id', 'function_id', 'started_at', 'ended_at', 'previous_id', 'status_id']

    def save(self, *args, **kwargs):
        if (
            self.status_id in (
                CollectingDataSubStageStatus.FAILED.key,
                CollectingDataSubStageStatus.READY_TO_EXPORT.key
            )
            and not self.ended_at
        ):
            self.ended_at = datetime.now()

        super().save(*args, **kwargs)


class ExportingDataStageStatus(TitledModelEnum):
    """
    Статус этапа выгрузки данных.
    """

    CREATED = ModelEnumValue(
        title='Создан',
    )

    IN_PROGRESS = ModelEnumValue(
        title='В процессе',
    )

    FAILED = ModelEnumValue(
        title='Завершено с ошибками',
    )

    FINISHED = ModelEnumValue(
        title='Завершено',
    )

    class Meta:
        db_table = 'rdm_exporting_data_stage_status'
        verbose_name = 'Модель-перечисление статусов этапа выгрузки данных'
        verbose_name_plural = 'Модели-перечисления статусов этапов выгрузки данных'


class ExportingDataStage(ReprStrPreModelMixin, BaseObjectModel):
    """
    Этап выгрузки данных.
    """

    manager = ForeignKey(
        to=Entity,
        verbose_name='Менеджер ранера Функции',
        on_delete=PROTECT,
        null=True,
        blank=True,
    )

    period_started_at = DateTimeField(
        'Левая граница периода выборки данных для выгрузки',
        db_index=True,
    )

    period_ended_at = DateTimeField(
        'Правая граница периода выборки данных для выгрузки',
        db_index=True,
    )

    started_at = DateTimeField(
        'Время начала выгрузки данных',
        auto_now_add=True,
    )

    ended_at = DateTimeField(
        'Время завершения выгрузки данных',
        null=True,
        blank=True,
    )

    status = ForeignKey(
        to=ExportingDataStageStatus,
        verbose_name='Статус',
        on_delete=PROTECT,
        default=ExportingDataStageStatus.CREATED.key,
    )

    class Meta:
        db_table = 'rdm_exporting_data_stage'
        verbose_name = 'Этап формирования данных для выгрузки'
        verbose_name_plural = 'Этапы формирования данных для выгрузки'

    @property
    def attrs_for_repr_str(self):
        return ['manager_id', 'started_at', 'ended_at', 'status_id']

    def save(self, *args, **kwargs):
        if (
            self.status_id in (ExportingDataStageStatus.FAILED.key, ExportingDataStageStatus.FINISHED.key)
            and not self.ended_at
        ):
            self.ended_at = datetime.now()

        super().save(*args, **kwargs)


class ExportingDataSubStageStatus(TitledModelEnum):
    """
    Модель-перечисление статусов этапа выгрузки данных
    """

    CREATED = ModelEnumValue(
        title='Создан',
    )

    IN_PROGRESS = ModelEnumValue(
        title='Запущен',
    )

    FAILED = ModelEnumValue(
        title='Завершено с ошибками',
    )

    FINISHED = ModelEnumValue(
        title='Завершен',
    )

    class Meta:
        db_table = 'rdm_exporting_data_sub_stage_status'
        verbose_name = 'Модель-перечисление статусов подэтапа выгрузки данных'
        verbose_name_plural = 'Модели-перечисления статусов подэтапов выгрузки данных'


class ExportingDataSubStage(ReprStrPreModelMixin, BaseObjectModel):
    """
    Подэтап выгрузки данных.
    """

    function = ForeignKey(
        to=Entity,
        verbose_name='Функция',
        on_delete=PROTECT,
        null=True,
        blank=True,
    )

    stage = ForeignKey(
        to=ExportingDataStage,
        verbose_name='Этап выгрузки данных',
        on_delete=CASCADE,
    )

    started_at = DateTimeField(
        verbose_name='Время начала сбора данных',
        auto_now_add=True,
        db_index=True,
    )

    ended_at = DateTimeField(
        verbose_name='Время завершения сбора данных',
        null=True,
        blank=True,
        db_index=True,
    )

    status = ForeignKey(
        to=ExportingDataSubStageStatus,
        verbose_name='Статус',
        on_delete=PROTECT,
        default=ExportingDataSubStageStatus.CREATED.key,
    )

    class Meta:
        db_table = 'rdm_exporting_data_sub_stage'
        verbose_name = 'Стадия выгрузки данных'
        verbose_name_plural = 'Стадии выгрузки данных'

    @property
    def attrs_for_repr_str(self):
        return ['function_id', 'collecting_data_sub_stage_id', 'stage_id', 'started_at', 'ended_at', 'status_id']

    def save(self, *args, **kwargs):
        if (
            self.status_id in (ExportingDataSubStageStatus.FAILED.key, ExportingDataSubStageStatus.FINISHED.key)
            and not self.ended_at
        ):
            self.ended_at = datetime.now()

        super().save(*args, **kwargs)


class ExportingDataSubStageAttachment(ReprStrPreModelMixin, BaseObjectModel):
    """
    Сгенерированный файл для дальнейшей выгрузки в "Региональная витрина данных".
    """

    exporting_data_sub_stage = ForeignKey(
        to=ExportingDataSubStage,
        verbose_name='Подэтап выгрузки данных',
        on_delete=CASCADE,
    )

    # TODO PYTD-22 В зависимости от принятого решения по инструменту ограничения доступа к media-файлам, нужно будет
    #  изменить тип поля или оставить как есть
    attachment = FileField(
        verbose_name='Сгенерированный файл',
        upload_to=get_exporting_data_stage_attachment_path,
        max_length=512,
        null=True,
        blank=True,
    )

    operation = SmallIntegerField(
        verbose_name='Действие',
        choices=EntityLogOperation.get_choices(),
    )

    created = DateTimeField(
        verbose_name='Дата создания',
        auto_now_add=True,
        null=True,
        blank=True,
    )
    modified = DateTimeField(
        verbose_name='Дата изменения',
        auto_now=True,
        null=True,
        blank=True,
    )

    class Meta:
        db_table = 'rdm_exporting_data_sub_stage_attachment'
        verbose_name = 'Сгенерированный файл для дальнейшей выгрузки в "Региональная витрина данных"'
        verbose_name_plural = 'Сгенерированные файлы для дальнейшей выгрузки в "Региональная витрина данных"'

    @property
    def attrs_for_repr_str(self):
        return ['exporting_data_sub_stage_id', 'attachment', 'operation', 'created', 'modified']


class DataMartRequestStatus(TitledIntegerModelEnum):
    """Модель-перечисление статусов загрузки данных в Витрину."""

    UPLOAD_TO_BUFFER = ModelEnumValue(
        value=-1,
        title='Загрузка данных в буффер',
    )

    BUFFERED = ModelEnumValue(
        value=0,
        title='Запрос буфферизирован',
    )

    WAIT_FOR_OPEN_DELTA = ModelEnumValue(
        value=1,
        title='Ожидает открытия дельты',
    )

    IN_PROCESSING = ModelEnumValue(
        value=2,
        title='В обработке',
    )

    SUCCESSFULLY_PROCESSED = ModelEnumValue(
        value=3,
        title='Успешно обработан',
    )

    FAILED_PROCESSING = ModelEnumValue(
        value=4,
        title='Ошибка обработки запроса',
    )

    REQUEST_ID_NOT_FOUND = ModelEnumValue(
        value=5,
        title='Идентификатор запроса не обнаружен',
    )

    FORMAT_LOGICAL_CONTROL = ModelEnumValue(
        value=6,
        title='Форматно-логический контроль',
    )

    FLC_ERROR = ModelEnumValue(
        value=7,
        title='Ошибки ФЛК',
    )

    class Meta:
        db_table = 'rdm_request_status'
        verbose_name = 'Статус загрузки данных в Витрину'
        verbose_name_plural = 'Статусы загрузки данных в Витрину'


class ExportingDataSubStageUploaderClientLog(ReprStrPreModelMixin, BaseObjectModel):
    """
    Связь лога Загрузчика данных с подэтапом выгрузки данных.
    """

    entry = ForeignKey(
        to=Entry,
        verbose_name='Лог запроса и ответа',
        on_delete=CASCADE,
        related_name='uploader_client_log',
    )

    sub_stage = ForeignKey(
        to=ExportingDataSubStage,
        verbose_name='Подэтап выгрузки данных',
        on_delete=CASCADE,
    )

    attachment = ForeignKey(
        to=ExportingDataSubStageAttachment,
        verbose_name='Прикрепленный файл',
        on_delete=CASCADE,
    )

    request_id = CharField(
        verbose_name='Id запроса загрузки в витрину',
        max_length=100,
        blank=True,
        db_index=True,
    )

    is_emulation = BooleanField(
        verbose_name='Включен режим эмуляции',
        default=False,
    )

    file_upload_status = SmallIntegerField(
        verbose_name="Общий статус загрузки файла в витрину",
        choices=FileUploadStatusEnum.get_choices(),
        null=True,
        blank=True,
    )

    created = DateTimeField(
        verbose_name='Дата создания',
        auto_now_add=True,
        null=True,
        blank=True,
        db_index=True,
    )
    modified = DateTimeField(
        verbose_name='Дата изменения',
        auto_now=True,
        null=True,
        blank=True,
        db_index=True,
    )

    class Meta:
        db_table = 'rdm_exporting_data_sub_stage_uploader_client_log'
        verbose_name = 'Лог запроса подэтапа выгрузки данных'
        verbose_name_plural = 'Лог запроса подэтапа выгрузки данных'


class UploadStatusRequestLog(ReprStrPreModelMixin, BaseObjectModel):
    """
    Модель связывающая статусы, загрузку файла в витрину и http-запрос к витрине.
    """

    upload = ForeignKey(
        verbose_name='Cвязь запроса статуса с загрузкой файла в витрину',
        to=ExportingDataSubStageUploaderClientLog,
        on_delete=CASCADE,
    )
    entry = ForeignKey(
        verbose_name='Cвязь запроса статуса с запросом в витрину',
        to=Entry,
        on_delete=CASCADE,
        related_name='upload_status_request_log',
    )
    request_status = ForeignKey(
        verbose_name='Статус запроса в витрине',
        to=DataMartRequestStatus,
        on_delete=PROTECT,
        null=True,
        blank=True,
    )

    class Meta:
        db_table = 'rdm_upload_status_request_log'
        verbose_name = 'Лог запроса подэтапа выгрузки данных'
        verbose_name_plural = 'Логи запроса подэтапа выгрузки данных'


class BaseEntityModel(ReprStrPreModelMixin, BaseObjectModel):
    """Базовая модель сущности."""

    collecting_sub_stage = ForeignKey(
        verbose_name='Подэтап сбора данных сущности',
        to=CollectingExportedDataSubStage,
        on_delete=CASCADE,
    )
    exporting_sub_stage = ForeignKey(
        verbose_name='Подэтап выгрузки',
        to=ExportingDataSubStage,
        blank=True,
        null=True,
        on_delete=CASCADE,
    )
    operation = SmallIntegerField(
        verbose_name='Действие',
        choices=EntityLogOperation.get_choices(),
    )

    created = DateTimeField(
        verbose_name='Дата создания',
        auto_now_add=True,
        null=True,
        blank=True,
        db_index=True,
    )
    modified = DateTimeField(
        verbose_name='Дата изменения',
        auto_now=True,
        null=True,
        blank=True,
        db_index=True,
    )

    @property
    def attrs_for_repr_str(self):
        return ['collecting_sub_stage', 'exporting_sub_stage', 'operation', 'created', 'modified']

    class Meta:
        abstract = True


class RegionalDataMartModelEnum(TitledModelEnum):
    """Модель-перечисление моделей "Региональная витрина данных"."""

    class Meta:
        db_table = 'rdm_model'
        extensible = True
        verbose_name = 'Модель-перечисление моделей "Региональной витрины данных"'
        verbose_name_plural = 'Модели-перечисления моделей "Региональной витрины данных"'


class RegionalDataMartEntityEnum(TitledModelEnum):
    """Модель-перечисление сущностей выгрузки в Региональная витрина данных.

    Поля:
        entity - data-класс сущности;
        main_model_enum - значение модели-перечисления RegionalDataMartModelEnum основной модели РВД для формирования
            данных сущности. Обычно это модель идентификаторы записей которой соответствуют идентификаторам в записях
            сущности. У этих записей проставляется подэтап выгрузки данных;
        additional_model_enums - кортеж значений модели-перечисления RegionalDataMartModelEnum с дополнительными
            моделями РВД, которые участвуют в формировании записей сущностей. Они показывают, что перед запуском
            экспорта данных сущности по ним и основной модели должен быть запущен сбор данных.
    """

    class Meta:
        db_table = 'rdm_entity'
        extensible = True
        verbose_name = 'Модель-перечисление сущностей "Региональной витрины данных"'
        verbose_name_plural = 'Модели-перечисления сущностей "Региональной витрины данных"'

    @property
    def model_enums(self):
        """Возвращает модели, от которых зависит сущность."""
        value = self.model_enum_value

        return self.get_model_enums_from_value(value=value)

    @staticmethod
    def get_model_enums_from_value(value: ModelEnumValue):
        """Получение значений модели-перечисления моделей по значению модели-перечисления сущностей."""
        model_enums = [value.main_model_enum, *value.additional_model_enums]

        return model_enums

    @classmethod
    def get_entities_model_enums(
        cls,
        entity_enums: List[ModelEnumValue],
        is_sorted: bool = True,
    ) -> List[ModelEnumValue]:
        """Получение списка значений модели-перечисления моделей RegionalDataMartModelEnum.

        Args:
            entity_enums: Список значений модели-перечисления сущностей RegionalDataMartEntityEnum;
            is_sorted: Необходимость сортировки значений модели-перечисления RegionalDataMartModelEnum по полю
                order_number.
        """
        model_enums = set()

        for entity_enum_value in entity_enums:
            model_enums.update(cls.get_model_enums_from_value(value=entity_enum_value))

        model_enums = list(model_enums)

        if is_sorted:
            model_enums = sorted(model_enums, key=lambda value: value.order_number)

        return model_enums


class AbstractCollectDataCommandProgress(ReprStrPreModelMixin, BaseObjectModel):
    """
    Модель, хранящая данные для формирования и отслеживания асинхронных задач по сбору данных.

    В реализации необходимо определить поля:
        1. Ссылку на асинхронную задачу, например:
            task = ForeignKey(
                to=RunningTask,
                verbose_name='Асинхронная задача',
                blank=True, null=True,
                on_delete=SET_NULL,
            )
        2. Поле хранения лога:
            logs_link = FileField(
                upload_to=upload_file_handler,
                max_length=255,
                verbose_name='Ссылка на файл логов',
            )
    """

    task = ...

    logs_link = ...

    stage = ForeignKey(
        to=CollectingExportedDataStage,
        verbose_name='Этап формирования данных для выгрузки',
        blank=True, null=True,
        on_delete=SET_NULL,
    )
    model = ForeignKey(
        to=RegionalDataMartModelEnum,
        verbose_name='Модель РВД',
        on_delete=PROTECT,
    )
    created = DateTimeField(
        verbose_name='Дата создания',
        default=django.utils.timezone.now,
    )
    logs_period_started_at = DateTimeField(
        'Левая граница периода обрабатываемых логов',
    )
    logs_period_ended_at = DateTimeField(
        'Правая граница периода обрабатываемых логов',
    )
    generation_id = UUIDField(
        'Идентификатор генерации',
        default=uuid.uuid4,
    )
    institute_ids = JSONField(
        'id учебного заведения',
        blank=True,
        null=True,
        default=list,
    )

    class Meta:
        abstract = True
        db_table = 'rdm_collecting_data_command_progress'
        verbose_name = 'Задача по сбору данных'
        verbose_name_plural = 'Задачи по сбору данных'


class UploaderClientLog(Entry):
    """Прокси модель Загрузчика данных в витрину."""

    objects = UploaderClientLogManager()
    base_objects = Manager()

    @cached_property
    def http_method_and_url(self) -> Tuple[str, str]:
        """Возвращает http-метод и url из поля запроса Entry.request."""
        request = StringIO(self.request)
        request_first_line = request.readline()
        request.close()

        method, url = request_first_line.split(' ')[:2]
        if not (method and url.startswith('http')):
            method = url = ''

        return method.strip('[]'), url

    @cached_property
    def http_response_status(self) -> str:
        """Статус-код запроса к витрине."""
        try:
            http_status = self.response.split(' ')[0].strip('[]')
        except IndexError:
            http_status = None
        return http_status

    @property
    def http_method(self) -> str:
        """Значение http-метода."""
        return self.http_method_and_url[0]

    @property
    def request_url(self) -> str:
        """URL запроса."""
        return self.http_method_and_url[1]

    @property
    def request_error(self) -> Optional[str]:
        """Ошибка запроса."""
        return self.error

    class Meta:
        proxy = True



from django.db.models import (
    Case,
    CharField,
    F,
    Manager,
    PositiveSmallIntegerField,
    Q,
    Value,
    When,
)
from django.db.models.functions import (
    Cast,
)

from edu_rdm_integration.uploader_log.enums import (
    RequestResultStatus,
)


class UploaderClientLogManager(Manager):
    """Менеджер модели журнала Загрузчика данных в витрину."""

    def get_queryset(self):
        """Возвращает кварисет."""
        query = super().get_queryset()

        result_status = Case(
            When(
                Q(Q(error__isnull=True) | Q(error__exact=''))
                & Q(Q(response__isnull=False) & ~Q(response__exact='')),
                then=Value(RequestResultStatus.SUCCESS)
            ),
            default=Value(RequestResultStatus.ERROR),
            output_field=PositiveSmallIntegerField(),
        )

        query = query.annotate(
            request_datetime=F('date_time'),
            attachment_file=Case(
                When(
                    request__icontains='POST',
                    then=F('uploader_client_log__attachment__attachment')
                ),
                When(
                    request__icontains='GET',
                    then=F('upload_status_request_log__upload__attachment__attachment'),
                ),
                default=Value('')
            ),
            status_code=Case(
                When(
                    request__icontains='POST',
                    then=Value('', output_field=CharField())
                ),
                When(
                    request__icontains='GET',
                    then=Cast(
                        'upload_status_request_log__request_status__value',
                        output_field=CharField()
                    ),
                ),
                default=Value('')
            ),
            status_description=Case(
                When(
                    request__icontains='POST',
                    then=Value('', output_field=CharField()),
                ),
                When(
                    request__icontains='GET',
                    then=F('upload_status_request_log__request_status__title'),
                ),
                default=Value(''), output_field=CharField()
            ),
            is_emulation=Case(
                When(
                    request__icontains='POST',
                    then=F('uploader_client_log__is_emulation'),
                ),
                When(
                    request__icontains='GET',
                    then=F('upload_status_request_log__upload__is_emulation'),
                ),
                default=Value(False)
            ),
            request_id=Case(
                When(
                    request__icontains='POST',
                    then=F('uploader_client_log__request_id')
                ),
                When(
                    request__icontains='GET',
                    then=F('upload_status_request_log__upload__request_id')
                ),
                default=Value('')
            ),
            result_status=result_status
        )

        return query



from django.apps import (
    AppConfig,
)


class UploaderLoggerConfig(AppConfig):  # noqa D101

    name = 'regional_data_mart_integration.uploader_log'
    label = 'regional_data_mart_integration_uploader_log'
    verbose_name = 'Журнал логов РВД'



from edu_rdm_integration.adapters.errors import (
    WebEduError,
)


class BaseExportDataError(WebEduError):
    """
    Базовая ошибка функций выгрузки данных для интеграции с "Региональная витрина данных".
    """



from educommon.audit_log.utils import (
    get_model_by_table,
)
from educommon.integration_entities.consts import (
    LOG_OPERATION_MAP,
)
from educommon.integration_entities.enums import (
    EntityLogOperation,
)

from edu_rdm_integration.collect_data.base.caches import (
    LogChange,
)
from edu_rdm_integration.mapping import (
    MODEL_FIELDS_LOG_FILTER,
)


class ReformatLogsMixin:
    """Миксин для преобразования логов к удобному для работы виду в кешах помощников функций."""

    def _reformat_logs(self):
        """
        Производится преобразование логов к удобному для работы виду.

        Предполагается вложенные словари. На первом уровне ключом будет название
        модели, на втором идентификатор записи.
        """
        for log in self.raw_logs:
            model = get_model_by_table(log.table)._meta.label

            if getattr(self, '_log_only_models', None) and (model not in self._log_only_models):
                # Пропускаем, если модель не входит в список отслеживаемых
                continue

            operation = LOG_OPERATION_MAP[log.operation]

            if operation in EntityLogOperation.values:
                fields = log.data
            else:
                fields = {}

            log_change = LogChange(
                operation=operation,
                fields=fields,
            )

            if not self._filter_log(model, log_change):
                # Если модель не отслеживается, то запись лога не сохраняем
                continue

            if log_change.operation == EntityLogOperation.DELETE:
                self.logs[model][log.object_id] = [log_change, ]
            else:
                self.logs[model][log.object_id].append(log_change)

    @staticmethod
    def _filter_log(model: str, log_change: LogChange) -> bool:
        """
        Производится проверка изменений на отслеживаемые поля.
        """
        is_filtered = False

        if model in MODEL_FIELDS_LOG_FILTER[log_change.operation]:
            filter_fields = MODEL_FIELDS_LOG_FILTER[log_change.operation][model]
            if filter_fields:
                # Если заданы конкретные поля, которые должны отслеживать
                for field in log_change.fields:
                    if field in filter_fields:
                        # Достаточно, чтобы хотя бы одно поле попало под фильтр
                        is_filtered = True
                        break
            else:
                # Модель отслеживается, но перечень фильтруемых полей не задан,
                # значит фильтруем все поля модели
                is_filtered = True

        return is_filtered



from typing import (
    Any,
    Dict,
    Optional,
    Tuple,
)

from django.db import (
    models,
)
from django.utils.decorators import (
    classproperty,
)

from m3_db_utils.mixins import (
    BaseEnumRegisterMixin,
)
from m3_db_utils.models import (
    ModelEnumValue,
)

from edu_rdm_integration.models import (
    RegionalDataMartEntityEnum,
    RegionalDataMartModelEnum,
)
from edu_rdm_integration.utils import (
    camel_to_underscore,
)


class EntityEnumRegisterMixin(BaseEnumRegisterMixin):
    """Миксин, для регистрации сущности в RegionalDataMartEntityEnum."""

    enum = RegionalDataMartEntityEnum
    """Модель-перечисление в которую регистрируется сущность."""

    main_model_enum: ModelEnumValue
    """Значение RegionalDataMartModelEnum, 
    основной модели РВД для формирования сущности."""

    additional_model_enums: Tuple[ModelEnumValue] = ()
    """Перечень дополнительных значений RegionalDataMartModelEnum, 
    которые участвуют в формировании записей сущностей"""

    title: str
    """Расшифровка сущности модели-перечисления"""

    @classproperty
    def key(cls) -> str:
        return camel_to_underscore(cls.__name__.rsplit('Entity', 1)[0], upper=True)

    @classmethod
    def get_register_params(cls) -> Dict[str, Any]:
        register_params = super().get_register_params()

        register_params['main_model_enum'] = getattr(cls, 'main_model_enum', None) or cls.get_main_model_enum()
        register_params['entity'] = cls
        register_params['additional_model_enums'] = cls.additional_model_enums or cls.get_additional_model_enums()

        return register_params

    # TODO EDUSCHL-20938 Удалить в рамках задачи
    @classmethod
    def get_main_model_enum(cls) -> Optional[ModelEnumValue]:
        """Возвращает значение модели перечисление основной модели сущности.

        В классе определяется поле main_model_enum или данный метод.

        !!! Временное решение. Будет удалено в рамках EDUSCHL-20938.
        """

    # TODO EDUSCHL-20938 Удалить в рамках задачи
    @classmethod
    def get_additional_model_enums(cls) -> Optional[Tuple[ModelEnumValue]]:
        """Возвращает кортеж значений модели-перечисления основной модели сущности.

        В классе определяется поле additional_model_enums или данный метод.

        !!! Временное решение. Будет удалено в рамках EDUSCHL-20938.
        """


class ModelEnumRegisterMixin(BaseEnumRegisterMixin):
    """Миксин, для регистрации модели в RegionalDataMartModelEnum."""

    enum = RegionalDataMartModelEnum
    """Модель-перечисление в которую регистрируется модель."""

    creating_trigger_models: Tuple[models.Model] = ()
    """Перечень моделей по которым генерируются логи."""

    loggable_models: Tuple[models.Model] = ()
    """Перечень моделей по которым собираются логи."""

    @classproperty
    def key(cls) -> str:
        return camel_to_underscore(cls.__name__).upper()

    @classproperty
    def title(cls):
        return cls._meta.verbose_name

    @classmethod
    def get_register_params(cls) -> Dict[str, Any]:
        register_params = super().get_register_params()
        register_params['model'] = cls
        register_params['creating_trigger_models'] = cls.creating_trigger_models
        register_params['loggable_models'] = cls.loggable_models

        return register_params



from typing import (
    Any,
    Dict,
    NamedTuple,
)

from educommon.integration_entities.enums import (
    EntityLogOperation,
)


class LogChange(NamedTuple):
    """Операция и значения измененных полей из лога."""

    operation: EntityLogOperation
    fields: Dict[str, Any]

    @property
    def is_create(self) -> bool:
        """Лог создания."""
        return self.operation == EntityLogOperation.CREATE

    @property
    def is_update(self) -> bool:
        """Лог изменения."""
        return self.operation == EntityLogOperation.UPDATE

    @property
    def is_delete(self) -> bool:
        """Лог удаления."""
        return self.operation == EntityLogOperation.DELETE
